"""Creates meshes on a 3D L-shape.
"""
import meshpy.tet
import numpy as np


def create_mesh(maxvol):
    # circumcirlce radius
    cc_radius = 10.0
    lx = 2.0 / np.sqrt(3.0) * cc_radius
    ly = lx
    lz = lx

    # create the mesh data structure
    # Corner points of the cube
    points = [
        (-0.5 * lx, -0.5 * ly, -0.5 * lz),
        (0.5 * lx, -0.5 * ly, -0.5 * lz),
        (0.5 * lx, 0.5 * ly, -0.5 * lz),
        (-0.5 * lx, 0.5 * ly, -0.5 * lz),
        (-0.5 * lx, -0.5 * ly, 0.5 * lz),
        (0.5 * lx, 0.5 * ly, 0.5 * lz),
        (-0.5 * lx, 0.5 * ly, 0.5 * lz),
        (0.0, -0.5 * ly, 0.5 * lz),
        (0.0, -0.5 * ly, 0.0),
        (0.5 * lx, -0.5 * ly, 0.0),
        (0.5 * lx, 0.0, 0.0),
        (0.5 * lx, 0.0, 0.5 * lz),
        (0.0, 0.0, 0.5 * lz),
        (0.0, 0.0, 0.0),
    ]
    facets = [
        [0, 1, 2, 3],
        [4, 7, 12, 11, 5, 6],
        [0, 1, 9, 8, 7, 4],
        [1, 2, 5, 11, 10, 9],
        [2, 5, 6, 3],
        [3, 6, 4, 0],
        [8, 13, 12, 7],
        [8, 9, 10, 13],
        [10, 11, 12, 13],
    ]
    # create the mesh
    # Set the geometry and build the mesh.
    info = meshpy.tet.MeshInfo()
    info.set_points(points)
    info.set_facets(facets)
    meshpy_mesh = meshpy.tet.build(info, max_volume=maxvol)

    return np.array(meshpy_mesh.points), np.array(meshpy_mesh.elements)


if __name__ == "__main__":
    import meshio

    points, cells = create_mesh(0.1)
    meshio.write("lshape3d.e", points, {"tetra": cells})



import numpy as np

import meshzoo

from .helpers import is_near_equal, signed_simplex_volumes


def test_up():
    x_range = np.linspace(0.0, 1.0, 11)
    y_range = np.linspace(0.0, 1.0, 11)
    points, cells = meshzoo.rectangle_tri(x_range, y_range, variant="up")
    assert len(points) == 121
    assert is_near_equal(np.sum(points, axis=0), [60.5, 60.5])
    assert len(cells) == 200
    assert np.all(signed_simplex_volumes(points, cells) > 0.0)


def test_up2():
    x_range = np.linspace(0.0, 1.0, 3)
    y_range = np.linspace(0.0, 1.0, 2)
    points, cells = meshzoo.rectangle_tri(x_range, y_range, variant="up")
    assert len(points) == 6
    assert is_near_equal(np.sum(points, axis=0), [3.0, 3.0])
    assert len(cells) == 4
    assert list(cells[0]) == [0, 1, 4]
    assert list(cells[2]) == [0, 4, 3]
    assert np.all(signed_simplex_volumes(points, cells) > 0.0)


def test_zigzag():
    x_range = np.linspace(0.0, 1.0, 11)
    y_range = np.linspace(0.0, 1.0, 11)
    points, cells = meshzoo.rectangle_tri(x_range, y_range, variant="zigzag")
    assert len(points) == 121
    assert is_near_equal(np.sum(points, axis=0), [60.5, 60.5])
    assert len(cells) == 200
    assert np.all(signed_simplex_volumes(points, cells) > 0.0)


def test_zigzag2():
    x_range = np.linspace(0.0, 1.0, 2)
    y_range = np.linspace(0.0, 1.0, 2)
    points, cells = meshzoo.rectangle_tri(x_range, y_range, variant="zigzag")
    assert len(points) == 4
    assert is_near_equal(np.sum(points, axis=0), [2.0, 2.0])
    assert len(cells) == 2
    assert np.all(signed_simplex_volumes(points, cells) > 0.0)


def test_zigzag3():
    x_range = np.linspace(0.0, 1.0, 3)
    y_range = np.linspace(0.0, 1.0, 2)
    points, cells = meshzoo.rectangle_tri(x_range, y_range, variant="zigzag")
    assert len(points) == 6
    assert is_near_equal(np.sum(points, axis=0), [3.0, 3.0])
    assert len(cells) == 4
    assert list(cells[0]) == [0, 1, 4]
    assert list(cells[2]) == [1, 2, 4]
    assert np.all(signed_simplex_volumes(points, cells) > 0.0)


def test_down():
    x_range = np.linspace(0.0, 1.0, 5)
    y_range = np.linspace(0.0, 1.0, 4)
    points, cells = meshzoo.rectangle_tri(x_range, y_range, variant="down")
    assert len(points) == 20
    assert len(cells) == 24
    assert np.all(signed_simplex_volumes(points, cells) > 0.0)
    # meshzoo.show2d(points, cells)


def test_center():
    x_range = np.linspace(0.0, 1.0, 11)
    y_range = np.linspace(0.0, 1.0, 9)
    points, cells = meshzoo.rectangle_tri(x_range, y_range, variant="center")
    meshzoo.show2d(points, cells)
    assert len(points) == 99
    assert len(cells) == 160
    assert np.all(signed_simplex_volumes(points, cells) > 0.0)


def test_quad1():
    x_range = np.linspace(0.0, 1.0, 11)
    y_range = np.linspace(0.0, 1.0, 11)
    points, cells = meshzoo.rectangle_quad(x_range, y_range)
    assert len(points) == 121
    assert is_near_equal(np.sum(points, axis=0), [60.5, 60.5])
    assert len(cells) == 100
    # meshzoo.save2d("rectangle-quad.svg", points, cells)


def test_quad2():
    x_range = np.linspace(0.0, 1.0, 3)
    y_range = np.linspace(0.0, 1.0, 2)
    points, cells = meshzoo.rectangle_quad(x_range, y_range)
    assert len(points) == 6
    assert is_near_equal(np.sum(points, axis=0), [3.0, 3.0])
    assert np.all(cells == [[0, 1, 4, 3], [1, 2, 5, 4]])


def test_quad3():
    x_range = np.linspace(0.0, 1.0, 3)
    y_range = np.linspace(0.0, 1.0, 3)
    points, cells = meshzoo.rectangle_quad(x_range, y_range)
    assert len(points) == 9
    assert np.all(cells == [[0, 1, 4, 3], [3, 4, 7, 6], [1, 2, 5, 4], [4, 5, 8, 7]])


def test_quad8():
    x_range = np.linspace(0.0, 1.0, 3)
    y_range = np.linspace(0.0, 1.0, 2)
    points, cells = meshzoo.rectangle_quad(x_range, y_range, cell_type="quad8")
    assert len(points) == 13
    print(cells.tolist())
    assert np.all(cells == [[0, 1, 4, 3, 6, 11, 8, 10], [1, 2, 5, 4, 7, 12, 9, 11]])


def test_quad9():
    x_range = np.linspace(0.0, 1.0, 4)
    y_range = np.linspace(0.0, 1.0, 3)
    points, cells = meshzoo.rectangle_quad(x_range, y_range, cell_type="quad9")
    assert len(points) == 35
    print(cells.tolist())
    assert np.all(
        cells
        == [
            [0, 1, 5, 4, 12, 22, 15, 21, 29],
            [4, 5, 9, 8, 15, 26, 18, 25, 32],
            [1, 2, 6, 5, 13, 23, 16, 22, 30],
            [5, 6, 10, 9, 16, 27, 19, 26, 33],
            [2, 3, 7, 6, 14, 24, 17, 23, 31],
            [6, 7, 11, 10, 17, 28, 20, 27, 34],
        ]
    )



import numpy as np


def create_edges(cells_nodes):
    """Setup edge-node and edge-cell relations. Adapted from voropy."""
    # Create the idx_hierarchy (nodes->edges->cells), i.e., the value of
    # `self.idx_hierarchy[0, 2, 27]` is the index of the node of cell 27, edge
    # 2, node 0. The shape of `self.idx_hierarchy` is `(2, 3, n)`, where `n` is
    # the number of cells. Make sure that the k-th edge is opposite of the k-th
    # point in the triangle.
    local_idx = np.array([[1, 2], [2, 0], [0, 1]]).T
    # Map idx back to the nodes. This is useful if quantities which are in
    # idx shape need to be added up into nodes (e.g., equation system rhs).
    nds = cells_nodes.T
    idx_hierarchy = nds[local_idx]

    s = idx_hierarchy.shape
    a = np.sort(idx_hierarchy.reshape(s[0], s[1] * s[2]).T)

    b = np.ascontiguousarray(a).view(np.dtype((np.void, a.dtype.itemsize * a.shape[1])))
    _, idx, inv, cts = np.unique(
        b, return_index=True, return_inverse=True, return_counts=True
    )

    # No edge has more than 2 cells. This assertion fails, for example, if
    # cells are listed twice.
    assert all(cts < 3)

    edge_nodes = a[idx]
    cells_edges = inv.reshape(3, -1).T

    return edge_nodes, cells_edges


def show2d(*args, **kwargs):
    import matplotlib.pyplot as plt

    plot2d(*args, **kwargs)
    plt.show()


def save2d(filename, *args, **kwargs):
    import matplotlib.pyplot as plt

    plot2d(*args, **kwargs)
    plt.savefig(filename, transparent=True, bbox_inches="tight")


def plot2d(
    points,
    cells,
    show_axes=False,
    # ParaView's default colors
    fill: str = "#c8c5bd",
    stroke: str = "#000080",
):
    """Plot a 2D mesh using matplotlib."""
    import matplotlib.pyplot as plt

    fig = plt.figure()
    ax = fig.gca()
    plt.axis("equal")
    if not show_axes:
        ax.set_axis_off()

    assert points.shape[1] == 2

    xmin = np.amin(points[:, 0])
    xmax = np.amax(points[:, 0])
    ymin = np.amin(points[:, 1])
    ymax = np.amax(points[:, 1])

    width = xmax - xmin
    xmin -= 0.1 * width
    xmax += 0.1 * width

    height = ymax - ymin
    ymin -= 0.1 * height
    ymax += 0.1 * height

    ax.set_xlim(xmin, xmax)
    ax.set_ylim(ymin, ymax)

    for cell in cells:
        import matplotlib.patches

        poly = matplotlib.patches.Polygon(points[cell], ec=stroke, fc=fill)
        ax.add_patch(poly)

    # import matplotlib.tri
    # tri = matplotlib.tri.Triangulation(points[:,0], points[:,1], triangles=cells)
    # ax.triplot(tri, '-', lw=1, color="k")
    return fig


def _compose_from_faces(corners, faces, n, edge_adjust=None, face_adjust=None):
    # create corner nodes
    vertices = [corners]
    vertex_count = len(corners)
    corner_nodes = np.arange(len(corners))

    # create edges
    edges = set()
    for face in faces:
        edges.add(tuple(sorted([face[0], face[1]])))
        edges.add(tuple(sorted([face[1], face[2]])))
        edges.add(tuple(sorted([face[2], face[0]])))

    edges = list(edges)

    # create edge nodes:
    edge_nodes = {}
    t = np.linspace(1 / n, 1.0, n - 1, endpoint=False)
    corners = vertices[0]
    k = corners.shape[0]
    for edge in edges:
        i0, i1 = edge
        new_vertices = np.outer(1 - t, corners[i0]) + np.outer(t, corners[i1])
        if edge_adjust:
            new_vertices = edge_adjust(edge, new_vertices)
        vertices.append(new_vertices)
        vertex_count += len(vertices[-1])
        edge_nodes[edge] = np.arange(k, k + len(t))
        k += len(t)

    # This is the same code as appearing for cell in a single triangle. On each face,
    # those indices are translated into the actual indices.
    triangle_cells = []
    k = 0
    for i in range(n):
        j = np.arange(n - i)
        triangle_cells.append(np.column_stack([k + j, k + j + 1, k + n - i + j + 1]))
        j = j[:-1]
        triangle_cells.append(
            np.column_stack([k + j + 1, k + n - i + j + 2, k + n - i + j + 1])
        )
        k += n - i + 1
    triangle_cells = np.vstack(triangle_cells)

    cells = []
    for face in faces:
        corners = face
        edges = [(face[0], face[1]), (face[1], face[2]), (face[2], face[0])]
        is_edge_reverted = [False, False, False]
        for k, edge in enumerate(edges):
            if edge[0] > edge[1]:
                edges[k] = (edge[1], edge[0])
                is_edge_reverted[k] = True

        # First create the interior points in barycentric coordinates
        if n == 1:
            num_new_vertices = 0
        else:
            bary = (
                np.hstack(
                    [[np.full(n - i - 1, i), np.arange(1, n - i)] for i in range(1, n)]
                )
                / n
            )
            bary = np.array([1.0 - bary[0] - bary[1], bary[1], bary[0]])
            corner_verts = np.array([vertices[0][i] for i in corners])
            vertices_cart = np.dot(corner_verts.T, bary).T

            if face_adjust:
                vertices_cart = face_adjust(face, bary, vertices_cart, corner_verts)

            vertices.append(vertices_cart)
            num_new_vertices = len(vertices[-1])

        # translation table
        num_nodes_per_triangle = (n + 1) * (n + 2) // 2
        tt = np.empty(num_nodes_per_triangle, dtype=int)

        # first the corners
        tt[0] = corner_nodes[corners[0]]
        tt[n] = corner_nodes[corners[1]]
        tt[num_nodes_per_triangle - 1] = corner_nodes[corners[2]]
        # then the edges.
        # edge 0
        tt[1:n] = edge_nodes[edges[0]]
        if is_edge_reverted[0]:
            tt[1:n] = tt[1:n][::-1]
        #
        # edge 1
        idx = 2 * n
        for k in range(n - 1):
            if is_edge_reverted[1]:
                tt[idx] = edge_nodes[edges[1]][n - 2 - k]
            else:
                tt[idx] = edge_nodes[edges[1]][k]
            idx += n - k - 1
        #
        # edge 2
        idx = n + 1
        for k in range(n - 1):
            if is_edge_reverted[2]:
                tt[idx] = edge_nodes[edges[2]][k]
            else:
                tt[idx] = edge_nodes[edges[2]][n - 2 - k]
            idx += n - k

        # now the remaining interior nodes
        idx = n + 2
        j = vertex_count
        for k in range(n - 2):
            for _ in range(n - k - 2):
                tt[idx] = j
                j += 1
                idx += 1
            idx += 2

        cells += [tt[triangle_cells]]
        vertex_count += num_new_vertices

    vertices = np.concatenate(vertices)
    cells = np.concatenate(cells)

    return vertices, cells


def insert_midpoints_edges(points, cells, cell_type):
    """Collect all unique edges, calculate and return points including
    midpoints on edges as well as the extended cells array."""

    number_of_points = {"triangle": 3, "tetra": 4, "quad": 4, "hexahedron": 8}

    if cells.shape[1] != number_of_points[cell_type]:
        raise ValueError("Mismatch of cell type and shape of cells array.")

    if cell_type == "triangle":
        # k-th edge between cell points no. (ij[k])
        ij = [[0, 1], [1, 2], [2, 0]]

    elif cell_type == "tetra":
        # k-th edge between cell points no. (ij[k])
        ij = [[0, 1], [1, 2], [2, 0], [3, 0], [3, 1], [3, 2]]

    elif cell_type == "quad":
        # k-th edge between cell points no. (ij[k])
        ij = [[0, 1], [1, 2], [2, 3], [3, 0]]

    elif cell_type == "hexahedron":
        # k-th edge between cell points no. (ij[k])
        ij = [
            [0, 1],
            [1, 2],
            [2, 3],
            [3, 0],
            [4, 5],
            [5, 6],
            [6, 7],
            [7, 4],
            [0, 4],
            [1, 5],
            [2, 6],
            [3, 7],
        ]

    else:
        raise TypeError("Cell type not implemented.")

    # obtain edges from cells (contains duplicates)
    edges = cells[:, ij]

    # sort points of edges
    edges_sorted = np.sort(edges.reshape(-1, 2), axis=1)

    # obtain unique edges and inverse mapping
    edges_unique, inverse = np.unique(
        edges_sorted,
        return_index=False,
        return_inverse=True,
        return_counts=False,
        axis=0,
    )

    # calculate midpoints on edges as mean of edge-based pairs of points
    midpoints_on_edges = np.mean(points[edges_unique.T], axis=0)

    # create the additional cells array
    # add offset to point index for midpoints on edges
    cells_edges = inverse.reshape(len(cells), -1) + len(points)

    # vertical stack of points and horizontal stack of edges
    points_new = np.vstack((points, midpoints_on_edges))
    cells_new = np.hstack((cells, cells_edges))

    return points_new, cells_new



import numpy as np

from ._cube import cube_hexa, cube_tetra


def ball_hexa(n: int):
    a = 1 / np.sqrt(3)
    ls = np.linspace(-a, a, n + 1)
    nodes, cells = cube_hexa(ls, ls, ls)

    # Inflate the nodes towards the circle boundary.
    # Inflate each point such that the 2-norm of the new point is the max-norm of the
    # old.
    alpha = np.max(np.abs(nodes), axis=1)
    beta = np.linalg.norm(nodes, axis=1)
    idx = beta > 1.0e-13
    nodes[idx] = (nodes[idx].T * (alpha[idx] / beta[idx])).T

    return nodes, cells


def ball_tetra(n: int):
    a = 1 / np.sqrt(3)
    ls = np.linspace(-a, a, n + 1)
    nodes, cells = cube_tetra(ls, ls, ls)

    # Inflate the nodes towards the circle boundary.
    # Inflate each point such that the 2-norm of the new point is the max-norm of the
    # old.
    alpha = np.max(np.abs(nodes), axis=1)
    beta = np.linalg.norm(nodes, axis=1)
    idx = beta > 1.0e-13
    nodes[idx] = (nodes[idx].T * (alpha[idx] / beta[idx])).T

    return nodes, cells



import numpy as np

from ._helpers import _compose_from_faces


def ngon(p: int, n: int, offset: float = np.pi / 2):
    x = np.linspace(offset, offset + 2 * np.pi, p, endpoint=False)
    corners = np.vstack(
        [
            [[0.0, 0.0]],
            np.array([np.cos(x), np.sin(x)]).T,
        ]
    )
    faces = [(0, k + 1, k + 2) for k in range(p - 1)] + [(0, p, 1)]
    return _compose_from_faces(corners, faces, n)



import numpy as np

import meshzoo

from .helpers import is_near_equal


def _compute_cells_normals_dir(points, cells):
    pc = points[cells]
    midpoints = np.sum(pc, axis=1) / 3.0
    nrm = np.sqrt(np.einsum("ij,ij->i", midpoints, midpoints))
    normals = (midpoints.T / nrm).T
    cross = np.cross(
        pc[:, 1, :] - pc[:, 0, :],
        pc[:, 2, :] - pc[:, 0, :],
    )
    return np.einsum("ij,ij->i", normals, cross)


def test_uv_sphere(num_points_per_circle=20, num_circles=10):
    points, cells = meshzoo.uv_sphere(num_points_per_circle, num_circles)
    assert len(points) == 162
    assert is_near_equal(np.sum(points, axis=0), [0.0, 0.0, 0.0])
    assert len(cells) == 320
    assert (_compute_cells_normals_dir(points, cells) > 0.0).all()
    assert np.all(np.abs(np.einsum("ij,ij->i", points, points) - 1.0) < 1.0e-10)


def test_geo_sphere(num_points_per_circle=20, num_circles=10):
    points, tri, quad = meshzoo.geo_sphere(num_points_per_circle, num_circles)
    # import meshio
    # meshio.write_points_cells("geo-sphere.vtk", points, {"triangle": tri, "quad": quad})
    assert len(points) == 162
    assert is_near_equal(np.sum(points, axis=0), [0.0, 0.0, 0.0])
    assert len(tri) == 40
    assert len(quad) == 140
    # assert (_compute_cells_normals_dir(points, cells) > 0.0).all()
    assert np.all(np.abs(np.einsum("ij,ij->i", points, points) - 1.0) < 1.0e-10)


def test_tetra_sphere(n=16):
    points, cells = meshzoo.tetra_sphere(n)
    # import meshio
    # meshio.write_points_cells("out.vtk", points, {"triangle": cells})
    assert len(points) == 514
    assert is_near_equal(np.sum(points, axis=0), [0.0, 0.0, 0.0])
    assert len(cells) == 1024
    assert (_compute_cells_normals_dir(points, cells) > 0.0).all()
    assert np.all(np.abs(np.einsum("ij,ij->i", points, points) - 1.0) < 1.0e-10)


def test_octa_sphere(n=16):
    points, cells = meshzoo.octa_sphere(n)
    assert len(points) == 1026
    assert is_near_equal(np.sum(points, axis=0), [0.0, 0.0, 0.0])
    assert len(cells) == 2048
    assert (_compute_cells_normals_dir(points, cells) > 0.0).all()
    assert np.all(np.abs(np.einsum("ij,ij->i", points, points) - 1.0) < 1.0e-10)


def test_icosa_sphere(n=16):
    points, cells = meshzoo.icosa_sphere(n)
    # import meshio
    # meshio.write_points_cells("out.vtk", points, {"triangle": cells})
    assert len(points) == 2562
    assert is_near_equal(np.sum(points, axis=0), [0.0, 0.0, 0.0])
    assert len(cells) == 5120
    assert (_compute_cells_normals_dir(points, cells) > 0.0).all()
    assert np.all(np.abs(np.einsum("ij,ij->i", points, points) - 1.0) < 1.0e-10)


if __name__ == "__main__":
    test_geo_sphere()



import numpy as np

import meshzoo

from .helpers import is_near_equal, signed_simplex_volumes


def _get_points(bary):
    corners = np.array(
        [[0.0, -0.5 * np.sqrt(3.0), +0.5 * np.sqrt(3.0)], [1.0, -0.5, -0.5]]
    )
    return np.dot(corners, bary).T


def test_triangle():
    bary, cells = meshzoo.triangle(4)
    assert len(bary.T) == 15
    assert is_near_equal(np.sum(_get_points(bary), axis=0), [0.0, 0.0])
    assert len(cells) == 16

    # make sure the order of the nodes in each cell is counterclockwise
    corner_coords = np.array([[0.0, 1.0, 0.0], [0.0, 0.0, 1.0]])
    coords = np.dot(corner_coords, bary)
    assert np.all(signed_simplex_volumes(coords.T, cells) > 0.0)


def test_plot2d():
    bary, cells = meshzoo.triangle(4)
    meshzoo.show2d(_get_points(bary), cells)


def test_edges():
    _, cells = meshzoo.triangle(2)
    edges_nodes, edges_cells = meshzoo.create_edges(cells)
    assert np.all(
        edges_nodes
        == [[0, 1], [0, 3], [1, 2], [1, 3], [1, 4], [2, 4], [3, 4], [3, 5], [4, 5]]
    )
    assert np.all(edges_cells == [[3, 1, 0], [5, 4, 2], [6, 3, 4], [8, 7, 6]])


if __name__ == "__main__":
    points, cells = meshzoo.triangle(5000)
    # import meshio
    # meshio.write_points_cells('triangle.vtk', points, {'triangle': cells})



import numpy as np


def tube(length: float = 1.0, radius: float = 1.0, n: int = 30):
    # Number of points along the width of the strip (>= 2)
    # Choose it such that we have approximately square boxes.
    nw = int(round(length * n / (2 * np.pi * radius)))

    # Generate suitable ranges for parametrization
    u_range = np.linspace(0.0, 2 * np.pi, num=n, endpoint=False)
    v_range = np.linspace(-0.5 * length, 0.5 * length, num=nw)

    # Create the vertices.
    proto_points = np.dstack(np.meshgrid(u_range, v_range, indexing="ij")).reshape(
        -1, 2
    )
    points = np.column_stack(
        [
            radius * np.cos(proto_points[:, 0]),
            radius * np.sin(proto_points[:, 0]),
            proto_points[:, 1],
        ]
    )

    # create the elements (cells)
    cells = []
    for i in range(n - 1):
        for j in range(nw - 1):
            cells.append([i * nw + j, (i + 1) * nw + j + 1, i * nw + j + 1])
            cells.append([i * nw + j, (i + 1) * nw + j, (i + 1) * nw + j + 1])

    # close the geometry
    for j in range(nw - 1):
        cells.append([(n - 1) * nw + j, j + 1, (n - 1) * nw + j + 1])
        cells.append([(n - 1) * nw + j, j, j + 1])

    return points, np.array(cells)



#!/usr/bin/env python
"""
Create irregular mesh on a regular tetrahedron centered at the origin.
"""
import meshpy.tet
import numpy as np


def create_tetrahedron_mesh(maxvol=0.1):
    # circumcircle radius
    r = 5.0

    # boundary points
    points = []
    points.append((0.0, 0.0, r))
    # theta = arccos(-1/3) (tetrahedral angle)
    costheta = -1.0 / 3.0
    sintheta = 2.0 / 3.0 * np.sqrt(2.0)
    # phi = 0.0
    sinphi = 0.0
    cosphi = 1.0
    points.append((r * cosphi * sintheta, r * sinphi * sintheta, r * costheta))
    # phi = np.pi * 2.0 / 3.0
    sinphi = np.sqrt(3.0) / 2.0
    cosphi = -0.5
    points.append((r * cosphi * sintheta, r * sinphi * sintheta, r * costheta))
    # phi = - np.pi * 2.0 / 3.0
    sinphi = -np.sqrt(3.0) / 2.0
    cosphi = -0.5
    points.append((r * cosphi * sintheta, r * sinphi * sintheta, r * costheta))

    # boundary faces
    facets = [[0, 1, 2], [0, 2, 3], [0, 3, 1], [1, 2, 3]]

    # create the mesh
    # Set the geometry and build the mesh.
    info = meshpy.tet.MeshInfo()
    info.set_points(points)
    info.set_facets(facets)
    meshpy_mesh = meshpy.tet.build(info, max_volume=maxvol)

    return np.array(meshpy_mesh.points), np.array(meshpy_mesh.elements)


if __name__ == "__main__":
    import meshio

    points, cells = create_tetrahedron_mesh(10.0)
    meshio.write("tetrahedron.vtu", points, {"tetra": cells})



import meshzoo

from .helpers import signed_simplex_volumes


def test_ball_hexa():
    points, cells = meshzoo.ball_hexa(10)
    # import meshio
    # meshio.Mesh(points, {"hexahedron": cells}).write("ball-hexa.vtk")
    assert len(points) == 1331
    assert len(cells) == 1000


def test_ball_tetra():
    points, cells = meshzoo.ball_tetra(10)
    # import meshio
    # meshio.Mesh(points, {"tetra": cells}).write("ball-tetra.vtk")
    assert len(points) == 1331
    assert len(cells) == 5000


def test_positive_volumes():
    points, cells = meshzoo.ball_tetra(2)
    assert (signed_simplex_volumes(points, cells) > 0.0).all()



import numpy as np


def moebius(
    num_twists: int = 1,  # How many twists are there in the 'paper'?
    nl: int = 60,  # Number of nodes along the length of the strip
    nw: int = 11,  # Number of nodes along the width of the strip (>= 2)
    variant: str = "classical",
):
    """Creates a simplistic triangular mesh on a slightly Möbius strip. The
    Möbius strip here deviates slightly from the ordinary geometry in that it
    is constructed in such a way that the two halves can be exchanged as to
    allow better comparison with the pseudo-Möbius geometry.

    The variant is either `'classical'` or `'smooth'`. The first is the classical
    Möbius band parametrization, the latter a smoothed variant matching
    `'pseudo'`.
    """
    # The width of the strip
    width = 1.0
    scale = 10.0

    # radius of the strip when flattened out
    r = 1.0

    # seam displacement
    alpha0 = 0.0  # pi / 2

    # How flat the strip will be.
    # Positive values result in left-turning Möbius strips, negative in
    # right-turning ones.
    # Also influences the width of the strip.
    flatness = 1.0

    # Generate suitable ranges for parametrization
    u_range = np.linspace(0.0, 2 * np.pi, num=nl, endpoint=False)
    v_range = np.linspace(-0.5 * width, 0.5 * width, num=nw)

    # Create the vertices. This is based on the parameterization
    # of the Möbius strip as given in
    # <https://en.wikipedia.org/wiki/M%C3%B6bius_strip#Geometry_and_topology>
    sin_u = np.sin(u_range)
    cos_u = np.cos(u_range)
    alpha = num_twists * 0.5 * u_range + alpha0
    sin_alpha = np.sin(alpha)
    cos_alpha = np.cos(alpha)

    if variant == "classical":
        a = cos_alpha
        b = sin_alpha
        reverse_seam = num_twists % 2 == 1
    elif variant == "smooth":
        # The fundamental difference with the ordinary Möbius band here are the
        # squares.
        # It is also possible to to abs() the respective sines and cosines, but
        # this results in a non-smooth manifold.
        a = np.copysign(cos_alpha ** 2, cos_alpha)
        b = np.copysign(sin_alpha ** 2, sin_alpha)
        reverse_seam = num_twists % 2 == 1
    else:
        assert variant == "pseudo"
        a = cos_alpha ** 2
        b = sin_alpha ** 2
        reverse_seam = False

    nodes = (
        scale
        * np.array(
            [
                np.outer(a * cos_u, v_range) + r * cos_u[:, np.newaxis],
                np.outer(a * sin_u, v_range) + r * sin_u[:, np.newaxis],
                np.outer(b, v_range) * flatness,
            ]
        )
        .reshape(3, -1)
        .T
    )

    cells = _create_elements(nl, nw, reverse_seam)
    return nodes, cells


def _create_elements(nl, nw, reverse_seam):
    cells = []
    for i in range(nl - 1):
        for j in range(nw - 1):
            if (i + j) % 2 == 0:
                cells.append([i * nw + j, (i + 1) * nw + j + 1, i * nw + j + 1])
                cells.append([i * nw + j, (i + 1) * nw + j, (i + 1) * nw + j + 1])
            else:
                cells.append([i * nw + j, i * nw + j + 1, (i + 1) * nw + j])
                cells.append([i * nw + j + 1, (i + 1) * nw + j, (i + 1) * nw + j + 1])

    # close the geometry
    i = nl - 1
    if reverse_seam:
        # Close the geometry upside down (odd Möbius fold)
        for j in range(nw - 1):
            if (i + j) % 2 == 0:
                cells.append([i * nw + j, (nw - 1) - (j + 1), i * nw + j + 1])
                cells.append([i * nw + j, (nw - 1) - j, (nw - 1) - (j + 1)])
            else:
                cells.append([i * nw + j, i * nw + j + 1, (nw - 1) - j])
                cells.append([i * nw + j + 1, (nw - 1) - j, (nw - 1) - (j + 1)])
    else:
        # Close the geometry upside up (even Möbius fold)
        for j in range(nw - 1):
            if (i + j) % 2 == 0:
                cells.append([i * nw + j, j + 1, i * nw + j + 1])
                cells.append([i * nw + j, j, j + 1])
            else:
                cells.append([i * nw + j, i * nw + j + 1, j])
                cells.append([i * nw + j + 1, j, j + 1])

    return np.array(cells)



import numpy as np

import meshzoo

from .helpers import is_near_equal


def test_tube():
    points, cells = meshzoo.tube(n=10)
    assert len(points) == 20
    assert is_near_equal(np.sum(points, axis=0), [0.0, 0.0, 0.0])
    assert len(cells) == 20



import numpy as np

from ._helpers import _compose_from_faces
from ._rectangle import rectangle_quad


def disk(p: int, n: int, offset: float = np.pi / 2):
    x = np.linspace(offset, offset + 2 * np.pi, p, endpoint=False)
    corners = np.vstack(
        [
            [[0.0, 0.0]],
            np.array([np.cos(x), np.sin(x)]).T,
        ]
    )
    faces = [(0, k + 1, k + 2) for k in range(p - 1)] + [(0, p, 1)]

    def edge_adjust(edge, verts):
        if 0 in edge:
            return verts
        dist = np.sqrt(np.einsum("ij,ij->i", verts, verts))
        return verts / dist[:, None]

    def face_adjust(face, bary, verts, corner_verts):
        assert face[0] == 0
        z = np.zeros_like(bary[1])
        edge_proj_bary = np.array([z, bary[1], bary[2]]) / (bary[1] + bary[2])
        edge_proj_cart = np.dot(corner_verts.T, edge_proj_bary).T
        dist = np.sqrt(np.einsum("ij,ij->i", edge_proj_cart, edge_proj_cart))
        return verts / dist[:, None]

    return _compose_from_faces(
        corners, faces, n, edge_adjust=edge_adjust, face_adjust=face_adjust
    )


def disk_quad(n: int, cell_type: str = "quad4"):
    a = 1 / np.sqrt(2)
    nodes, cells = rectangle_quad(
        np.linspace(-a, a, n + 1), np.linspace(-a, a, n + 1), cell_type
    )

    # Inflate the nodes towards the circle boundary.
    # Inflate each point such that the 2-norm of the new point is the max-norm of the
    # old.
    alpha = np.max(np.abs(nodes), axis=1)
    beta = np.linalg.norm(nodes, axis=1)
    idx = beta > 1.0e-13
    nodes[idx] = (nodes[idx].T * (alpha[idx] / beta[idx])).T

    return nodes, cells



import numpy as np
from meshpy.geometry import EXT_OPEN, GeometryBuilder, generate_surface_of_revolution
from meshpy.tet import MeshInfo, build


def create_ball_mesh(num_longi_points=10):

    radius = 5.0

    radial_subdiv = 2 * num_longi_points

    dphi = np.pi / num_longi_points

    # Make sure the nodes meet at the poles of the ball.
    def truncate(r):
        if abs(r) < 1e-10:
            return 0
        else:
            return r

    # Compute the volume of a canonical tetrahedron
    # with edgelength radius*dphi.
    a = radius * dphi
    canonical_tet_volume = np.sqrt(2.0) / 12 * a ** 3

    # Build outline for surface of revolution.
    rz = [
        (truncate(radius * np.sin(i * dphi)), radius * np.cos(i * dphi))
        for i in range(num_longi_points + 1)
    ]

    geob = GeometryBuilder()
    geob.add_geometry(
        *generate_surface_of_revolution(
            rz, closure=EXT_OPEN, radial_subdiv=radial_subdiv
        )
    )
    mesh_info = MeshInfo()
    geob.set(mesh_info)
    meshpy_mesh = build(mesh_info, max_volume=canonical_tet_volume)

    return np.array(meshpy_mesh.points), np.array(meshpy_mesh.elements)


if __name__ == "__main__":
    import meshio

    points, cells = create_ball_mesh()
    meshio.write("ball.e", points, {"tetra": cells})



#!/usr/bin/env python

import meshpy.triangle
import numpy as np
from scipy import special


def create_mesh(axis0=1, axis1=0.5, num_boundary_points=100):
    # lengths of major and minor axes
    a = max(axis0, axis1)
    b = min(axis0, axis1)

    # Choose the maximum area of a triangle equal to the area of
    # an equilateral triangle on the boundary.
    # For circumference of an ellipse, see
    # https://en.wikipedia.org/wiki/Ellipse#Circumference
    eccentricity = np.sqrt(1.0 - (b / a) ** 2)
    length_boundary = float(4 * a * special.ellipe(eccentricity))
    a_boundary = length_boundary / num_boundary_points
    max_area = a_boundary ** 2 * np.sqrt(3) / 4

    # generate points on the circle
    Phi = np.linspace(0, 2 * np.pi, num_boundary_points, endpoint=False)
    boundary_points = np.column_stack((a * np.cos(Phi), b * np.sin(Phi)))

    info = meshpy.triangle.MeshInfo()
    info.set_points(boundary_points)

    def _round_trip_connect(start, end):
        result = []
        for i in range(start, end):
            result.append((i, i + 1))
        result.append((end, start))
        return result

    info.set_facets(_round_trip_connect(0, len(boundary_points) - 1))

    def _needs_refinement(vertices, area):
        return bool(area > max_area)

    meshpy_mesh = meshpy.triangle.build(info, refinement_func=_needs_refinement)

    # append column
    pts = np.array(meshpy_mesh.points)
    points = np.c_[pts[:, 0], pts[:, 1], np.zeros(len(pts))]

    return points, np.array(meshpy_mesh.elements)


if __name__ == "__main__":
    import meshio

    points, cells = create_mesh()
    meshio.write("ellipse.e", points, {"triangle": cells})



import ball
import cube
import ellipse
import lshape
import lshape3d
import pacman
import rectangle
import rectangle_with_hole
import tetrahedron
import torus

__all__ = [
    "ball",
    "cube",
    "ellipse",
    "lshape",
    "lshape3d",
    "pacman",
    "rectangle",
    "rectangle_with_hole",
    "tetrahedron",
    "torus",
]



import dolfin
import mshr


def create_toy_mesh():

    box = mshr.Box(dolfin.Point(-3, -1, -0.5), dolfin.Point(3, 1, 0.5))

    c1 = mshr.Cylinder(dolfin.Point(0, 0, -2), dolfin.Point(0, 0, 2), 0.6, 0.6)
    b1 = mshr.Box(dolfin.Point(-2.5, -0.5, -2), dolfin.Point(-1.5, 0.5, 2))

    # "triangle"
    t1 = mshr.Polygon(
        [
            dolfin.Point(2.5, -0.5, 0),
            dolfin.Point(2.5, 0.5, 0),
            dolfin.Point(1.5, -0.5, 0),
        ]
    )
    g3d = mshr.Extrude2D(t1, -2)
    g3d = mshr.CSGTranslation(g3d, dolfin.Point(0, 0, 1))

    m = mshr.generate_mesh(box - c1 - b1 - g3d, 40, "cgal")

    return m.coordinates(), m.cells()


if __name__ == "__main__":
    import meshio

    points, cells = create_toy_mesh()
    meshio.write("toy.e", points, {"tetra": cells})



from ._ball import ball_hexa, ball_tetra
from ._cube import cube, cube_hexa, cube_tetra
from ._disk import disk, disk_quad
from ._helpers import create_edges, insert_midpoints_edges, plot2d, save2d, show2d
from ._moebius import moebius
from ._ngon import ngon
from ._rectangle import rectangle, rectangle_quad, rectangle_tri
from ._sphere import geo_sphere, icosa_sphere, octa_sphere, tetra_sphere, uv_sphere
from ._triangle import triangle
from ._tube import tube

__all__ = [
    "ball_hexa",
    "ball_tetra",
    "cube",
    "cube_tetra",
    "cube_hexa",
    "disk",
    "disk_quad",
    "moebius",
    "ngon",
    "rectangle",
    "rectangle_tri",
    "rectangle_quad",
    "uv_sphere",
    "icosa_sphere",
    "octa_sphere",
    "tetra_sphere",
    "geo_sphere",
    "triangle",
    "tube",
    #
    "save2d",
    "show2d",
    "plot2d",
    "create_edges",
    "insert_midpoints_edges",
]



"""
Creates a mesh for a circle with a cut.
"""
import meshpy.triangle
import numpy as np


def create_pacman_mesh(num_boundary_points=50):
    n_phi = num_boundary_points
    radius = 5.0

    # set those to 0.0 for perfect circle
    cut_angle = 0.1 * 2 * np.pi
    cut_depth = 0.5 * radius

    # Choose the maximum area of a triangle equal to the area of
    # an equilateral triangle on the boundary.
    a_boundary = (2 * np.pi - cut_angle) * radius / n_phi
    max_area = a_boundary ** 2 * np.sqrt(3.0) / 4.0
    max_area = float(max_area)  # meshpy can't deal with np.float64

    # generate points on the boundary
    Phi = np.linspace(
        0.5 * cut_angle, 2 * np.pi - 0.5 * cut_angle, n_phi, endpoint=False
    )
    boundary_points = []
    if abs(cut_angle) > 0.0 or cut_depth != 0.0:
        boundary_points.append((radius - cut_depth, 0.0))
    for phi in Phi:
        boundary_points.append((radius * np.cos(phi), radius * np.sin(phi)))

    # create the mesh
    info = meshpy.triangle.MeshInfo()
    info.set_points(boundary_points)

    def _round_trip_connect(start, end):
        result = []
        for i in range(start, end):
            result.append((i, i + 1))
        result.append((end, start))
        return result

    info.set_facets(_round_trip_connect(0, len(boundary_points) - 1))

    def _needs_refinement(vertices, area):
        return bool(area > max_area)

    meshpy_mesh = meshpy.triangle.build(info, refinement_func=_needs_refinement)

    # append column
    pts = np.array(meshpy_mesh.points)
    points = np.c_[pts[:, 0], pts[:, 1], np.zeros(len(pts))]

    return points, np.array(meshpy_mesh.elements)


if __name__ == "__main__":
    import meshio

    points, cells = create_pacman_mesh()
    meshio.write("pacman.vtu", points, {"triangle": cells})



import numpy as np

import meshzoo

from .helpers import signed_simplex_volumes


def test_ngon():
    points, cells = meshzoo.ngon(9, 8)
    # meshzoo.save2d("9gon.svg", points, cells)
    assert len(points) == 325
    assert len(cells) == 576
    assert np.all(signed_simplex_volumes(points, cells) > 0.0)


if __name__ == "__main__":
    test_ngon()



import numpy as np
import pytest

import meshzoo

from .helpers import is_near_equal


@pytest.mark.parametrize(
    "num_twists, num_points, num_cells, ref1, ref2",
    [
        [1, 5890, 11400, [0, 0, 0], [2753575 / 9.0, 2724125 / 9.0, 58900 / 3.0]],
        [2, 5890, 11400, [0, 0, 0], [2797750 / 9.0, 2679950 / 9.0, 58900 / 3.0]],
    ],
)
def test_moebius(num_twists, num_points, num_cells, ref1, ref2):
    points, cells = meshzoo.moebius(num_twists, 190, 31, variant="smooth")
    assert len(points) == num_points
    assert len(cells) == num_cells
    assert is_near_equal(np.sum(points, axis=0), ref1, tol=1.0e-10)
    sum_points2 = np.sum(points ** 2, axis=0)
    assert np.allclose(sum_points2, ref2, rtol=1.0e-12, atol=0.0)


@pytest.mark.parametrize(
    "num_twists, num_points, num_cells, ref1, ref2",
    [
        [
            1,
            5700,
            11020,
            [0, 0, 0],
            [[296107.21982759, 292933.72844828, 19040.94827586]],
        ],
        [
            2,
            5700,
            11020,
            [0, 0, 0],
            [[300867.45689655, 288173.49137931, 19040.94827586]],
        ],
    ],
)
def test_moebius2(num_twists, num_points, num_cells, ref1, ref2):
    points, cells = meshzoo.moebius(
        nl=190, nw=30, num_twists=num_twists, variant="smooth"
    )
    assert len(points) == num_points
    assert len(cells) == num_cells
    assert is_near_equal(np.sum(points, axis=0), ref1, tol=1.0e-10)
    sum_points2 = np.sum(points ** 2, axis=0)
    assert np.allclose(sum_points2, ref2, rtol=1.0e-12, atol=0.0)


@pytest.mark.parametrize(
    "num_twists, num_points, num_cells, ref1, ref2",
    [
        [1, 1000, 1800, [0, 0, 0], [1418750 / 27.0, 1418750 / 27.0, 137500 / 27.0]],
        [2, 1000, 1800, [0, 0, 0], [484375 / 9.0, 1384375 / 27.0, 137500 / 27.0]],
    ],
)
def test_moebius3(num_twists, num_points, num_cells, ref1, ref2):
    points, cells = meshzoo.moebius(num_twists, 100, 10, variant="classical")
    assert len(points) == num_points
    assert len(cells) == num_cells
    assert is_near_equal(np.sum(points, axis=0), ref1, tol=1.0e-10)
    sum_points2 = np.sum(points ** 2, axis=0)
    assert np.allclose(sum_points2, ref2, rtol=1.0e-12, atol=0.0)


def test_pseudomoebius():
    points, cells = meshzoo.moebius(nl=190, nw=31, variant="pseudo")
    assert len(points) == 5890
    assert len(cells) == 11400
    assert is_near_equal(np.sum(points, axis=0), [0, 0, 0], tol=1.0e-10)
    sum_points2 = np.sum(points ** 2, axis=0)
    ref2 = [2753575 / 9.0, 2724125 / 9.0, 58900 / 3.0]
    assert np.allclose(sum_points2, ref2, rtol=1.0e-12, atol=0.0)



import numpy as np


def triangle(n: int):
    # Create the mesh in barycentric coordinates
    bary = np.hstack(
        [
            [
                np.full(n - i + 1, i / n),
                # np.arange(n - i + 1) / n:
                np.linspace(0.0, (n - i + 1) / n, n - i + 1, endpoint=False),
            ]
            for i in range(n + 1)
        ]
    )
    bary = np.array([1.0 - bary[0] - bary[1], bary[1], bary[0]])

    # Some applications don't accept values like -1.4125e-16.
    bary[bary < 0.0] = 0.0
    bary[bary > 1.0] = 1.0

    cells = []
    k = 0
    for i in range(n):
        j = np.arange(n - i)
        cells.append(np.column_stack([k + j, k + j + 1, k + n - i + j + 1]))
        #
        j = j[:-1]
        cells.append(np.column_stack([k + j + 1, k + n - i + j + 2, k + n - i + j + 1]))
        k += n - i + 1

    cells = np.vstack(cells)
    return bary, cells



import meshzoo

from .helpers import signed_simplex_volumes


def test_disk():
    points, cells = meshzoo.disk(9, 8)
    meshzoo.save2d("4gon_disk.svg", points, cells)
    assert len(points) == 325
    assert len(cells) == 576
    assert (signed_simplex_volumes(points, cells) > 0.0).all()


def test_disk_quad():
    points, cells = meshzoo.disk_quad(10)
    # meshzoo.save2d("disk-quad.svg", points, cells)
    assert len(points) == 121
    assert len(cells) == 100



from __future__ import annotations

import numpy as np
from numpy.typing import ArrayLike


# backwards compatibility
def cube(
    x0: float,
    x1: float,
    y0: float,
    y1: float,
    z0: float,
    z1: float,
    nx: int,
    ny: int,
    nz: int,
):
    return cube_tetra(
        np.linspace(x0, x1, nx + 1),
        np.linspace(y0, y1, ny + 1),
        np.linspace(z0, z1, nz + 1),
    )


def cube_hexa(
    x_range: ArrayLike, y_range: ArrayLike, z_range: ArrayLike
) -> tuple[np.ndarray, np.ndarray]:
    x_range = np.asarray(x_range)
    y_range = np.asarray(y_range)
    z_range = np.asarray(z_range)

    nx1 = len(x_range)
    ny1 = len(y_range)
    nz1 = len(z_range)

    # Create the vertices.
    x, y, z = np.meshgrid(x_range, y_range, z_range, indexing="ij")

    # Alternative with slightly different order:
    # ```
    # points = np.stack([x, y, z]).T.reshape(-1, 3)
    # ```
    points = np.array([x, y, z]).T.reshape(-1, 3)

    # Create the cells.
    a = np.arange(len(points)).reshape(nz1, ny1, nx1)
    a = np.transpose(a, [2, 1, 0])

    # `c` contains the indices of each "cube" like
    #
    #
    #          c[7]     c[6]
    #            ________
    #           /       /|
    #     c[4] /_______/ | c[5]
    #         |        | |
    #         |        | |
    #         |        | / c[2]
    #         |________|/
    #
    #       c[0]      c[1]
    #
    cells = (
        np.array(
            [
                a[:-1, :-1, :-1],
                a[1:, :-1, :-1],
                a[1:, 1:, :-1],
                a[:-1, 1:, :-1],
                a[:-1, :-1, 1:],
                a[1:, :-1, 1:],
                a[1:, 1:, 1:],
                a[:-1, 1:, 1:],
            ]
        )
        .reshape(8, -1)
        .T
    )

    return points, cells


def cube_tetra(x_range: ArrayLike, y_range: ArrayLike, z_range: ArrayLike):
    x_range = np.asarray(x_range)
    y_range = np.asarray(y_range)
    z_range = np.asarray(z_range)

    nx1 = len(x_range)
    ny1 = len(y_range)
    nz1 = len(z_range)

    # Create the vertices.
    x, y, z = np.meshgrid(x_range, y_range, z_range, indexing="ij")
    # Alternative with slightly different order:
    # ```
    # points = np.stack([x, y, z]).T.reshape(-1, 3)
    # ```
    points = np.array([x, y, z]).T.reshape(-1, 3)

    # Create the elements (cells).
    a = np.arange(len(points)).reshape(nz1, ny1, nx1)
    a = np.transpose(a, [2, 1, 0])

    #
    # `c` contains the indices of each "cube" like
    #
    #
    #          c[7]     c[6]
    #            ________
    #           /       /|
    #     c[4] /_______/ | c[5]
    #         |        | |
    #         |        | |
    #         |        | / c[2]
    #         |________|/
    #
    #       c[0]      c[1]
    #
    c = [
        a[:-1, :-1, :-1],
        a[1:, :-1, :-1],
        a[1:, 1:, :-1],
        a[:-1, 1:, :-1],
        a[:-1, :-1, 1:],
        a[1:, :-1, 1:],
        a[1:, 1:, 1:],
        a[:-1, 1:, 1:],
    ]

    # 3D checkers
    idx = np.ones((nx1 - 1, ny1 - 1, nz1 - 1), dtype=bool)
    idx[::2, 1::2, ::2] = False
    idx[::2, ::2, 1::2] = False
    idx[1::2, ::2, ::2] = False
    idx[1::2, 1::2, 1::2] = False

    # There is 1 way to split a cube into 5 tetrahedra,
    # and 12 ways to split it into 6 tetrahedra.
    # See
    # <http://www.baumanneduard.ch/Splitting%20a%20cube%20in%20tetrahedras2.htm>
    # Also interesting: <https://en.wikipedia.org/wiki/Marching_tetrahedrons>.
    cells = [
        # regular; make sure the order of the points is such that the signed volume is
        # positive
        [c[0][idx], c[1][idx], c[3][idx], c[4][idx]],
        [c[1][idx], c[2][idx], c[3][idx], c[6][idx]],
        [c[1][idx], c[3][idx], c[4][idx], c[6][idx]],
        [c[1][idx], c[4][idx], c[5][idx], c[6][idx]],
        [c[3][idx], c[4][idx], c[6][idx], c[7][idx]],
        # the rest rotated such that it fits with the others; basically we change
        # "bottom" and "top" of the dice
        [c[4][~idx], c[5][~idx], c[0][~idx], c[7][~idx]],
        [c[5][~idx], c[6][~idx], c[2][~idx], c[7][~idx]],
        [c[5][~idx], c[7][~idx], c[2][~idx], c[0][~idx]],
        [c[5][~idx], c[0][~idx], c[2][~idx], c[1][~idx]],
        [c[7][~idx], c[0][~idx], c[3][~idx], c[2][~idx]],
    ]

    cells = np.column_stack([np.array(c).reshape(4, -1) for c in cells]).T

    return points, cells



import numpy as np

from ._helpers import _compose_from_faces


def uv_sphere(num_points_per_circle: int, num_circles: int, radius: float = 1.0):
    # Mesh parameters
    n_phi = num_points_per_circle
    n_theta = num_circles

    # Generate suitable ranges for parametrization
    phi_range = np.linspace(0.0, 2 * np.pi, num=n_phi, endpoint=False)
    theta_range = np.linspace(
        -np.pi / 2 + np.pi / (n_theta - 1),
        np.pi / 2 - np.pi / (n_theta - 1),
        num=n_theta - 2,
    )

    # nodes in the circles of latitude (except poles)
    nodes = radius * np.array(
        [[0.0, 0.0, -1.0]]  # south pole
        + [
            [
                np.cos(theta) * np.sin(phi),
                np.cos(theta) * np.cos(phi),
                np.sin(theta),
            ]
            for theta in theta_range
            for phi in phi_range
        ]
        + [[0.0, 0.0, 1.0]]  # north pole
    )

    south_pole_index = 0
    north_pole_index = len(nodes) - 1

    # create the elements (cells)
    num_cells = 2 * (n_theta - 2) * n_phi
    cells = []

    # connections to south pole
    for i in range(n_phi - 1):
        cells.append([south_pole_index, i + 1, i + 2])
    # close geometry
    cells.append([south_pole_index, n_phi, 1])

    # non-pole elements
    for i in range(n_theta - 3):
        for j in range(n_phi - 1):
            cells += [
                [i * n_phi + j + 2, i * n_phi + j + 1, (i + 1) * n_phi + j + 2],
                [i * n_phi + j + 1, (i + 1) * n_phi + j + 1, (i + 1) * n_phi + j + 2],
            ]

    # close the geometry
    for i in range(n_theta - 3):
        cells += [
            [i * n_phi + 1, (i + 1) * n_phi, (i + 1) * n_phi + 1],
            [(i + 1) * n_phi + 1, (i + 1) * n_phi, (i + 2) * n_phi],
        ]

    # connections to the north pole
    for i in range(n_phi - 1):
        cells.append(
            [
                i + 1 + n_phi * (n_theta - 3) + 1,
                i + n_phi * (n_theta - 3) + 1,
                north_pole_index,
            ]
        )
    # close geometry
    cells.append(
        [
            0 + n_phi * (n_theta - 3) + 1,
            n_phi - 1 + n_phi * (n_theta - 3) + 1,
            north_pole_index,
        ]
    )
    cells = np.array(cells)
    assert len(cells) == num_cells, "Wrong element count."

    return nodes, cells


def geo_sphere(num_points_per_circle: int, num_circles: int, radius=1.0):
    # Mesh parameters
    n_phi = num_points_per_circle
    n_theta = num_circles

    # Generate suitable ranges for parametrization
    phi_range = np.linspace(0.0, 2 * np.pi, num=n_phi, endpoint=False)
    theta_range = np.linspace(
        -np.pi / 2 + np.pi / (n_theta - 1),
        np.pi / 2 - np.pi / (n_theta - 1),
        num=n_theta - 2,
    )

    # nodes in the circles of latitude (except poles)
    nodes = radius * np.array(
        [[0.0, 0.0, -1.0]]  # south pole
        + [
            [
                np.cos(theta) * np.sin(phi),
                np.cos(theta) * np.cos(phi),
                np.sin(theta),
            ]
            for theta in theta_range
            for phi in phi_range
        ]
        + [[0.0, 0.0, 1.0]]  # north pole
    )

    south_pole_index = 0
    north_pole_index = len(nodes) - 1

    # create the elements (cells)
    tri = []
    quad = []

    # connections to south pole
    for i in range(n_phi - 1):
        tri.append([south_pole_index, i + 1, i + 2])
    # close geometry
    tri.append([south_pole_index, n_phi, 1])

    # non-pole elements
    for i in range(n_theta - 3):
        for j in range(n_phi - 1):
            quad += [
                [
                    i * n_phi + j + 1,
                    i * n_phi + j + 2,
                    (i + 1) * n_phi + j + 2,
                    (i + 1) * n_phi + j + 1,
                ],
            ]

    # close the geometry
    for i in range(n_theta - 3):
        quad += [
            [i * n_phi + 1, (i + 1) * n_phi + 1, (i + 2) * n_phi, (i + 1) * n_phi],
        ]

    # connections to the north pole
    for i in range(n_phi - 1):
        tri.append(
            [
                i + 1 + n_phi * (n_theta - 3) + 1,
                i + n_phi * (n_theta - 3) + 1,
                north_pole_index,
            ]
        )
    # close geometry
    tri.append(
        [
            0 + n_phi * (n_theta - 3) + 1,
            n_phi - 1 + n_phi * (n_theta - 3) + 1,
            north_pole_index,
        ]
    )
    tri = np.array(tri)
    quad = np.array(quad)

    return nodes, tri, quad


def tetra_sphere(n):
    corners = np.array(
        [
            [2 * np.sqrt(2) / 3, 0.0, -1.0 / 3.0],
            [-np.sqrt(2) / 3, np.sqrt(2.0 / 3.0), -1.0 / 3.0],
            [-np.sqrt(2) / 3, -np.sqrt(2.0 / 3.0), -1.0 / 3.0],
            [0.0, 0.0, 1.0],
        ]
    )
    # make sure the normals are pointing outwards
    faces = [(0, 2, 1), (0, 1, 3), (0, 3, 2), (1, 2, 3)]

    vertices, cells = _compose_from_faces(corners, faces, n)

    # push all nodes to the sphere
    norms = np.sqrt(np.einsum("ij,ij->i", vertices, vertices))
    vertices = (vertices.T / norms).T

    return vertices, cells


def octa_sphere(n: int):
    corners = np.array(
        [
            [1.0, 0.0, 0.0],
            [-1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, -1.0, 0.0],
            [0.0, 0.0, 1.0],
            [0.0, 0.0, -1.0],
        ]
    )
    faces = [
        (0, 2, 4),
        (1, 4, 2),
        (1, 3, 4),
        (0, 4, 3),
        (0, 5, 2),
        (1, 2, 5),
        (1, 5, 3),
        (0, 3, 5),
    ]
    vertices, cells = _compose_from_faces(corners, faces, n)

    # push all nodes to the sphere
    norms = np.sqrt(np.einsum("ij,ij->i", vertices, vertices))
    vertices = (vertices.T / norms).T

    return vertices, cells


def icosa_sphere(n: int):
    assert n >= 1
    # Start off with an isosahedron and refine.

    # Construction from
    # <http://blog.andreaskahler.com/2009/06/creating-icosphere-mesh-in-code.html>.
    # Create 12 vertices of a icosahedron.
    t = (1.0 + np.sqrt(5.0)) / 2.0
    corners = np.array(
        [
            [-1, +t, +0],
            [+1, +t, +0],
            [-1, -t, +0],
            [+1, -t, +0],
            #
            [+0, -1, +t],
            [+0, +1, +t],
            [+0, -1, -t],
            [+0, +1, -t],
            #
            [+t, +0, -1],
            [+t, +0, +1],
            [-t, +0, -1],
            [-t, +0, +1],
        ]
    )

    faces = [
        (0, 11, 5),
        (0, 5, 1),
        (0, 1, 7),
        (0, 7, 10),
        (0, 10, 11),
        (1, 5, 9),
        (5, 11, 4),
        (11, 10, 2),
        (10, 7, 6),
        (7, 1, 8),
        (3, 9, 4),
        (3, 4, 2),
        (3, 2, 6),
        (3, 6, 8),
        (3, 8, 9),
        (4, 9, 5),
        (2, 4, 11),
        (6, 2, 10),
        (8, 6, 7),
        (9, 8, 1),
    ]

    vertices, cells = _compose_from_faces(corners, faces, n)
    # push all nodes to the sphere
    norms = np.sqrt(np.einsum("ij,ij->i", vertices, vertices))
    vertices = (vertices.T / norms).T

    return vertices, cells



from __future__ import annotations

import numpy as np
from numpy.typing import ArrayLike


# backwards compatibility
def rectangle(
    x0: float,
    x1: float,
    y0: float,
    y1: float,
    nx: int,
    ny: int,
):
    x_range = np.linspace(x0, x1, nx + 1)
    y_range = np.linspace(y0, y1, ny + 1)
    return rectangle_tri(x_range, y_range)


def rectangle_quad(x_range: ArrayLike, y_range: ArrayLike, cell_type: str = "quad4"):
    x_range = np.asarray(x_range)
    y_range = np.asarray(y_range)

    nx = len(x_range)
    ny = len(y_range)

    if cell_type == "quad4":
        points = np.array(np.meshgrid(x_range, y_range)).reshape(2, -1).T
        a = np.arange(nx * ny).reshape(ny, nx).T

        cells = (
            np.array([a[:-1, :-1], a[1:, :-1], a[1:, 1:], a[:-1, 1:]]).reshape(4, -1).T
        )
    elif cell_type == "quad8":
        k = 0
        points_corner = np.array(np.meshgrid(x_range, y_range)).reshape(2, -1).T
        a_corner = np.arange(nx * ny).reshape(ny, nx).T
        k += a_corner.size

        x_mid = (x_range[:-1] + x_range[1:]) / 2
        points_xmid = np.array(np.meshgrid(x_mid, y_range)).reshape(2, -1).T
        a_xmid = k + np.arange((nx - 1) * ny).reshape(ny, nx - 1).T
        k += a_xmid.size

        y_mid = (y_range[:-1] + y_range[1:]) / 2
        points_ymid = np.array(np.meshgrid(x_range, y_mid)).reshape(2, -1).T
        a_ymid = k + np.arange(nx * (ny - 1)).reshape(ny - 1, nx).T

        points = np.row_stack([points_corner, points_xmid, points_ymid])

        cells = (
            np.array(
                [
                    a_corner[:-1, :-1],
                    a_corner[1:, :-1],
                    a_corner[1:, 1:],
                    a_corner[:-1, 1:],
                    a_xmid[:, :-1],
                    a_ymid[1:, :],
                    a_xmid[:, 1:],
                    a_ymid[:-1, :],
                ]
            )
            .reshape(8, -1)
            .T
        )
    else:
        assert cell_type == "quad9"
        k = 0
        points_corner = np.array(np.meshgrid(x_range, y_range)).reshape(2, -1).T
        a_corner = np.arange(nx * ny).reshape(ny, nx).T
        k += a_corner.size

        x_mid = (x_range[:-1] + x_range[1:]) / 2
        points_xmid = np.array(np.meshgrid(x_mid, y_range)).reshape(2, -1).T
        a_xmid = k + np.arange((nx - 1) * ny).reshape(ny, nx - 1).T
        k += a_xmid.size

        y_mid = (y_range[:-1] + y_range[1:]) / 2
        points_ymid = np.array(np.meshgrid(x_range, y_mid)).reshape(2, -1).T
        a_ymid = k + np.arange(nx * (ny - 1)).reshape(ny - 1, nx).T
        k += a_ymid.size

        points_center = np.array(np.meshgrid(x_mid, y_mid)).reshape(2, -1).T
        a_center = k + np.arange((nx - 1) * (ny - 1)).reshape(ny - 1, nx - 1).T

        points = np.row_stack([points_corner, points_xmid, points_ymid, points_center])

        cells = (
            np.array(
                [
                    a_corner[:-1, :-1],
                    a_corner[1:, :-1],
                    a_corner[1:, 1:],
                    a_corner[:-1, 1:],
                    a_xmid[:, :-1],
                    a_ymid[1:, :],
                    a_xmid[:, 1:],
                    a_ymid[:-1, :],
                    a_center,
                ]
            )
            .reshape(9, -1)
            .T
        )

    return points, cells


def rectangle_tri(x_range: ArrayLike, y_range: ArrayLike, variant: str = "zigzag"):
    x_range = np.asarray(x_range)
    y_range = np.asarray(y_range)

    nx = len(x_range)
    ny = len(y_range)

    # Create the vertices.
    points = np.array(np.meshgrid(x_range, y_range)).reshape(2, -1).T

    a = np.arange(nx * ny).reshape(ny, nx).T

    # indices of corners
    #
    # c[3]   c[2]
    #    _____
    #    |   |
    #    |___|
    #
    # c[0]   c[1]
    #
    c = [
        a[:-1, :-1],
        a[1:, :-1],
        a[1:, 1:],
        a[:-1, 1:],
    ]

    if variant == "up":
        cells = [
            [c[0], c[1], c[2]],
            [c[0], c[2], c[3]],
        ]
    elif variant == "down":
        cells = [
            [c[0], c[1], c[3]],
            [c[1], c[2], c[3]],
        ]
    elif variant == "zigzag":
        # https://stackoverflow.com/a/68550456/353337
        idx = np.ones((nx - 1, ny - 1), dtype=bool)
        idx[1::2, ::2] = False
        idx[::2, 1::2] = False
        cells = [
            # up
            [c[0][idx], c[1][idx], c[2][idx]],
            [c[0][idx], c[2][idx], c[3][idx]],
            # down
            [c[0][~idx], c[1][~idx], c[3][~idx]],
            [c[1][~idx], c[2][~idx], c[3][~idx]],
        ]
    else:
        assert variant == "center"
        idx = np.ones(((nx - 1), (ny - 1)), dtype=bool)
        idx[: (nx - 1) // 2, : (ny - 1) // 2] = False
        idx[(nx - 1) // 2 :, (ny - 1) // 2 :] = False

        cells = [
            # up
            [c[0][idx], c[1][idx], c[2][idx]],
            [c[0][idx], c[2][idx], c[3][idx]],
            # down
            [c[0][~idx], c[1][~idx], c[3][~idx]],
            [c[1][~idx], c[2][~idx], c[3][~idx]],
        ]

    cells = np.column_stack([np.array(c).reshape(3, -1) for c in cells]).T

    return points, cells



import math

import numpy as np


def is_near_equal(a, b, tol=1.0e-12):
    return np.allclose(a, b, rtol=0.0, atol=tol)


def signed_simplex_volumes(coords, cells):
    assert len(coords.shape) == 2
    assert len(cells.shape) == 2
    assert (
        coords.shape[1] + 1 == cells.shape[1]
    ), "Signed areas only make sense for n-simplices in in nD."

    # bc = coords[:, cells]
    # return np.cross((bc[:, :, 1] - bc[:, :, 0]).T, (bc[:, :, 2] - bc[:, :, 0]).T)

    # https://en.wikipedia.org/wiki/Simplex#Volume
    cp = coords[cells]
    n = coords.shape[1]
    # append ones; this appends a column instead of a row as suggested by
    # wikipedia, but that doesn't change the determinant
    cp1 = np.concatenate([cp, np.ones(cp.shape[:-1] + (1,))], axis=-1)

    sign = -1 if n % 2 == 1 else 1
    return sign * np.linalg.det(cp1) / math.factorial(n)



"""
Creates a mesh on a rectangle in the x-y-plane.
"""
import meshpy.triangle
import numpy as np


def create_mesh(edgelength=1.0, max_area=0.01):
    # dimensions of the rectangle
    lx = edgelength
    ly = edgelength

    # corner points
    boundary_points = [
        (-0.5 * lx, -0.5 * ly),
        (0.5 * lx, -0.5 * ly),
        (0.5 * lx, 0.5 * ly),
        (-0.5 * lx, 0.5 * ly),
    ]

    info = meshpy.triangle.MeshInfo()
    info.set_points(boundary_points)

    def _round_trip_connect(start, end):
        result = []
        for i in range(start, end):
            result.append((i, i + 1))
        result.append((end, start))
        return result

    info.set_facets(_round_trip_connect(0, len(boundary_points) - 1))

    def _needs_refinement(vertices, area):
        return bool(area > max_area)

    meshpy_mesh = meshpy.triangle.build(info, refinement_func=_needs_refinement)

    # append column
    pts = np.array(meshpy_mesh.points)
    points = np.c_[pts[:, 0], pts[:, 1], np.zeros(len(pts))]

    return points, np.array(meshpy_mesh.elements)


if __name__ == "__main__":
    import meshio

    points, cells = create_mesh()
    meshio.write("rectangle.e", points, {"triangle": cells})



import numpy as np

import meshzoo


def test_midpoints_edges_tri():
    points, cells = meshzoo.rectangle_tri(
        np.linspace(0.0, 1.0, 2), np.linspace(0.0, 1.0, 2)
    )

    assert len(points) == 4
    assert cells.shape == (2, 3)

    points_new, cells_new = meshzoo.insert_midpoints_edges(
        points, cells, cell_type="triangle"
    )

    assert len(points_new) == 9
    assert cells_new.shape == (2, 6)


def test_midpoints_edges_tetra():
    ls = np.linspace(0.0, 1.0, 2)
    points, cells = meshzoo.cube_tetra(ls, ls, ls)

    assert len(points) == 8
    assert cells.shape == (5, 4)

    points_new, cells_new = meshzoo.insert_midpoints_edges(
        points, cells, cell_type="tetra"
    )

    assert len(points_new) == 26
    assert cells_new.shape == (5, 10)


def test_midpoints_edges_quad():
    ls = np.linspace(0.0, 1.0, 3)
    points, cells = meshzoo.rectangle_quad(ls, ls)

    assert len(points) == 9
    assert cells.shape == (4, 4)

    points_new, cells_new = meshzoo.insert_midpoints_edges(
        points, cells, cell_type="quad"
    )

    assert len(points_new) == 21
    assert cells_new.shape == (4, 8)


def test_midpoints_edges_hexa():
    ls = np.linspace(0.0, 1.0, 3)
    points, cells = meshzoo.cube_hexa(ls, ls, ls)

    assert len(points) == 27
    assert cells.shape == (8, 8)

    points_new, cells_new = meshzoo.insert_midpoints_edges(
        points, cells, cell_type="hexahedron"
    )

    assert len(points_new) == 81
    assert cells_new.shape == (8, 20)


if __name__ == "__main__":
    test_midpoints_edges_tri()
    test_midpoints_edges_tetra()
    test_midpoints_edges_quad()
    test_midpoints_edges_hexa()



#!/usr/bin/env python

import numpy as np
from meshpy.geometry import (
    EXT_CLOSED_IN_RZ,
    GeometryBuilder,
    generate_surface_of_revolution,
)
from meshpy.tet import MeshInfo, build


def create_mesh(big_r=1.0, small_r=0.5, num_points=10):
    dphi = 2 * np.pi / num_points

    # Compute the volume of a canonical tetrahedron
    # with edgelength radius2*dphi.
    a = small_r * dphi
    canonical_tet_volume = np.sqrt(2.0) / 12 * a ** 3

    radial_subdiv = int(2 * np.pi * big_r / a)

    rz = [
        (big_r + small_r * np.cos(i * dphi), 0.5 * small_r * np.sin(i * dphi))
        for i in range(num_points)
    ]

    geob = GeometryBuilder()
    geob.add_geometry(
        *generate_surface_of_revolution(
            rz, closure=EXT_CLOSED_IN_RZ, radial_subdiv=radial_subdiv
        )
    )
    mesh_info = MeshInfo()
    geob.set(mesh_info)
    meshpy_mesh = build(mesh_info, max_volume=canonical_tet_volume)

    return np.array(meshpy_mesh.points), np.array(meshpy_mesh.elements)


if __name__ == "__main__":
    import meshio

    points, cells = create_mesh()
    meshio.write("torus.e", points, {"tetra": cells})
